--[[
Name: DogTag-1.0
Revision: $Rev: 33859 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Inspired By: WatchDog by Vika
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/wiki/DogTag-1.0
SVN: svn://svn.wowace.com/wowace/trunk/DogTag-1.0/DogTag-1.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: AceLibrary
License: LGPL v2.1
]]

-------------------------------------------------------------------
-- FOR TAG INFORMATION SEE http://www.wowace.com/wiki/DogTag-1.0 --
-------------------------------------------------------------------

local MAJOR_VERSION = "DogTag-1.0"
local MINOR_VERSION = "$Revision: 33859 $"

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local DogTag = {}

local L = {
	-- short races
	["BE"] = "BE",
	["Dr"] = "Dr",
	["Dw"] = "Dw",
	["Gn"] = "Gn",
	["Hu"] = "Hu",
	["NE"] = "NE",
	["Or"] = "Or",
	["Ta"] = "Ta",
	["Tr"] = "Tr",
	["Ud"] = "Ud",
	
	-- short classes
	["Wr"] = "Wr",
	["Pr"] = "Pr",
	["Ma"] = "Ma",
	["Sh"] = "Sh",
	["Pa"] = "Pa",
	["Wl"] = "Wl",
	["Dr"] = "Dr",
	["Ro"] = "Ro",
	["Hu"] = "Hu",
	
	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["r"] = "r",
	["r+"] = "r+",
	["+"] = "+",
	["b"] = "b",
	
	["Dead"] = DEAD,
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	
	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,
	
	["PvP"] = PVP,
	
	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,
	
	-- shortgenders
	["m"] = "m",
	["f"] = "f",
	
	["Rested"] = "Rested",
	
	["Leader"] = "Leader",
	["L"] = "L", -- short leader
}
if GetLocale() == "deDE" then
	-- alter L here
end

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.resting = { 1, 1, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }
colors.minHP = { 1, 0, 0 }
colors.midHP = { 1, 1, 0 }
colors.maxHP = { 0, 1, 0 }

local UnitCanAttack = UnitCanAttack
local UnitIsPVP = UnitIsPVP
local UnitPlayerControlled = UnitPlayerControlled
local UnitIsPlayer = UnitIsPlayer
local UnitClass = UnitClass
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitMana = UnitMana
local UnitManaMax = UnitManaMax
local UnitAffectingCombat = UnitAffectingCombat
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsDead = UnitIsDead
local UnitIsGhost = UnitIsGhost
local UnitIsConnected = UnitIsConnected
local IsResting = IsResting
local GetPetHappiness = GetPetHappiness
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitReaction = UnitReaction
local UnitLevel = UnitLevel
local GetDifficultyColor = GetDifficultyColor
local UnitPowerType = UnitPowerType
local UnitIsFriend = UnitIsFriend
local UnitCreatureFamily = UnitCreatureFamily
local UnitCreatureType = UnitCreatureType
local GetComboPoints = GetComboPoints
local UnitClassification = UnitClassification
local UnitFactionGroup = UnitFactionGroup
local UnitRace = UnitRace
local UnitIsPlusMob = UnitIsPlusMob
local UnitSex = UnitSex
local GetRestState = GetRestState
local UnitIsPartyLeader = UnitIsPartyLeader
local GetPetExperience = GetPetExperience
local UnitXP = UnitXP
local UnitXPMax = UnitXPMax
local GetXPExhaustion = GetXPExhaustion
local GetNumRaidMembers = GetNumRaidMembers
local GetRaidRosterInfo = GetRaidRosterInfo
local UnitName = UnitName
local GetGuildInfo = GetGuildInfo

local function GetClassColor(unit)
	local _,class = UnitClass(unit)
	return unpack(colors[class] or colors.unknown)
end

local function GetHostilityColor(unit)
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			-- they can attack me
			if UnitCanAttack("player", unit) then
				-- and I can attack them
				return unpack(colors.hostile)
			else
				-- but I can't attack them
				return unpack(colors.civilian)
			end
		elseif UnitCanAttack("player", unit) then
			-- they can't attack me, but I can attack them
			return unpack(colors.neutral)
		elseif UnitIsPVP(unit) then
			-- on my team
			return unpack(colors.friendly)
		else
			-- either enemy or friend, no violence
			return unpack(colors.civilian)
		end
	elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) or UnitIsDead(unit) then
		return unpack(colors.tapped)
	else
		local reaction = UnitReaction(unit, "player")
		if reaction then
			if reaction >= 5 then
				return unpack(colors.friendly)
			elseif reaction == 4 then
				return unpack(colors.neutral)
			else
				return unpack(colors.hostile)
			end
		else
			return unpack(colors.unknown)
		end
	end
end

local function HealthGradient(perc)
	local r1, g1, b1
	local r2, g2, b2
	if perc <= 0.5 then
		perc = perc * 2
		r1, g1, b1 = unpack(colors.minHP)
		r2, g2, b2 = unpack(colors.midHP)
	else
		perc = perc * 2 - 1
		r1, g1, b1 = unpack(colors.midHP)
		r2, g2, b2 = unpack(colors.maxHP)
	end
	return r1 + (r2-r1)*perc, g1 + (g2-g1)*perc, b1 + (b2-b1)*perc
end
-- Note, these tables use the unlocalized second returns of UnitClass and UnitRace, thus the keys are not localized.
local shortraces = {
	BloodElf = L["BE"],
	Draenei = L["Dr"],
	Dwarf = L["Dw"],
	Gnome = L["Gn"],
	Human = L["Hu"],
	NightElf = L["NE"],
	Orc = L["Or"],
	Tauren = L["Ta"],
	Troll = L["Tr"],
	Undead = L["Ud"],
}
local shortclasses = {
	WARRIOR = L["Wr"],
	PRIEST = L["Pr"],
	MAGE = L["Ma"],
	SHAMAN = L["Sh"],
	PALADIN = L["Pa"],
	WARLOCK = L["Wl"],
	DRUID = L["Dr"],
	ROGUE = L["Ro"],
	HUNTER = L["Hu"],
}

local Modifiers; Modifiers = {}
-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- note: can take an argument, either nil, a number, a string.
Modifiers.isequal = {function(unit, arg, value)
	if value == arg then
		return value
	end
end, value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.isless = {function(unit, arg, value)
	if type(value) == type(arg) then
		if value < arg then
			return value
		end
	else
		if tostring(value) < tostring(arg) then
			return value
		end
	end
end, value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.islessequal = {function(unit, arg, value)
	if type(value) == type(arg) then
		if value <= arg then
			return value
		end
	else
		if tostring(value) <= tostring(arg) then
			return value
		end
	end
end, value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.contains = {function(unit,arg,value)
	if string.find( value, arg ) then
		return value
	end
end, value = "string", arg = "string", ret = "string;nil"}

Modifiers.isgreater = "~islessequal"
Modifiers.isgreaterequal = "~isless"
Modifiers.isnotequal = "~isequal"

Modifiers.hide = "~isequal"
Modifiers.hideless = "~isless"
Modifiers.hidegreater = "~isgreater"
Modifiers.hidelessequal = "~islessequal"
Modifiers.hidegreaterequal = "~isgreaterequal"
Modifiers.hideunless = "~isnotequal"
Modifiers.hidezero = "hide(0)"

Modifiers.prepend = {function(unit, arg, value)
	return arg .. value
end, value = "number;string", arg = "string", ret = "string"}

Modifiers.append = {function(unit, arg, value)
	return value .. arg
end, value = "number;string", arg = "string", ret = "string"}

Modifiers.negate = {function(unit, value)
	return -value
end, value = "number", ret = "number"}

Modifiers.percent = {function(unit, value)
	return value .. "%"
end, value = "number", ret = "string"}

Modifiers.hidemaxhp = {function(unit, value)
	local maxValue = UnitHealthMax(unit)
	local currValue = UnitHealth(unit)
	if maxValue == 0 or currValue == maxValue then
		return
	else
		return value
	end
end, value = "number;string", ret = "same;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Modifiers.hidemaxmana = {function(unit, value)
	if UnitMana(unit) == UnitManaMax(unit) then
		return
	end
	return value
end, value = "number;string", ret = "same;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Modifiers.hidenomana = {function(unit, value)
	if UnitManaMax(unit) == 0 then
		return
	end
	return value
end, value = "number;string", ret = "same;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Modifiers.hidenocombat = {function(unit, value)
	if UnitAffectingCombat(unit) then
		return value
	end
end, value = "number;string", ret = "same;nil", events = "PLAYER_REGEN_ENABLED;PLAYER_REGEN_DISABLED;Update" }

Modifiers.hidegone = {function(unit, value)
	if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) then
		return nil
	end
	return value
end, value = "number;string", ret = "same;nil", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST" }

local function short(unit, value)
	if type(value) ~= "number" then
		if value:match("/") then
			local a,b = value:match("^(%d+)/(%d+)$")
			a = short(unit, tonumber(a))
			b = short(unit, tonumber(b))
			if a and b then
				return a.."/"..b
			end
		end
		return value
	end
	if value >= 10000000 or value <= -10000000 then
		return ("%.1fm"):format(value / 1000000)
	elseif value >= 1000000 or value <= -1000000 then
		return ("%.2fm"):format(value / 1000000)
	elseif value >= 100000 or value <= -100000 then
		return ("%.0fk"):format(value / 1000)
	elseif value >= 10000 or value <= -10000 then
		return ("%.1fk"):format(value / 1000)
	else
		return value
	end
end
Modifiers.short = {short,
	value = "number;string", ret = "same;nil"}

local function veryshort(unit, value)
	if type(value) ~= "number" then
		if value:match("/") then
			local a,b = value:match("^(%d+)/(%d+)$")
			a = veryshort(unit, tonumber(a))
			b = veryshort(unit, tonumber(b))
			return a.."/"..b
		end
		return value
	end
	if value >= 1000000 or value <= -1000000 then
		return ("%.0fm"):format(value / 1000000)
	elseif value >= 1000 or value <= -1000 then
		return ("%.0fk"):format(value / 1000)
	else
		return value
	end
end
Modifiers.veryshort = {veryshort,
	value = "number;string", ret = "same;nil"}

Modifiers.upper = {function(unit, value)
	return value:upper()
end, value = "string", ret = "string"}

Modifiers.lower = {function(unit, value)
	return value:lower()
end, value = "string", ret = "string"}

Modifiers.bracket = {function(unit, value)
	return ("[%s]"):format(value)
end, value = "number;string", ret = "string"}

Modifiers.paren = {function(unit, value)
	return ("(%s)"):format(value)
end, value = "number;string", ret = "string"}

Modifiers.angle = {function(unit, value)
	return ("<%s>"):format(value)
end, value = "number;string", ret = "string"}

Modifiers.brace = {function(unit, value)
	return ("{%s}"):format(value)
end, value = "number;string", ret = "string"}

Modifiers.trunc = {function(unit, arg, value)
	return tostring(value):sub(1, arg)
end, arg = "number", value = "string", ret = "string"}

Modifiers.contains = {function(unit, arg, value)
	if string.find(value, arg, 1, true) then
		return value
	end
end, arg = "string", value = "string", ret = "string;nil"}

Modifiers.hidecontains = "~contains"

Modifiers.statuscolor = {function(unit, value)
	local r,g,b
	if not UnitIsConnected(unit) then
		r,g,b = unpack(colors.disconnected)
	elseif UnitIsDeadOrGhost(unit) then
		r,g,b = unpack(colors.dead)
	elseif UnitAffectingCombat(unit) then
		r,g,b = unpack(colors.inCombat)
	elseif u == "player" and IsResting() then
		r,g,b = unpack(colors.resting)
	else
		if value then
			return ("|r%s"):format(value)
		else
			return "|r"
		end
	end	
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	else
		return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
	end
end, value = "number;string", ret = "string", events = "Update" --[[due to the combat thing]] }

Modifiers.happycolor = {function(unit, value)
	local x = GetPetHappiness()
	local r,g,b
	if x == 3 then
		r,g,b = unpack(colors.petHappy)
	elseif x == 2 then
		r,g,b = unpack(colors.petNeutral)
	elseif x == 1 then
		r,g,b = unpack(colors.petAngry)
	else
		return value
	end
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	else
		return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
	end
end, value = "number;string", ret = "string", events = "UNIT_HAPPINESS" }

Modifiers.hostilecolor = {function(unit, value)
	local r, g, b = GetHostilityColor(unit)
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	else
		return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
	end
end, value = "number;string", ret = "string", events = "UNIT_FACTION" }

Modifiers.classcolor = {function(unit, value)
	local r, g, b = GetClassColor(unit)
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	else
		return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
	end
end, value = "number;string", ret = "string"}

Modifiers.hpcolor = {function(unit, value)
	local perc = UnitHealth(unit) / UnitHealthMax(unit)
	local r, g, b = HealthGradient(perc)
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
	else
		return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
	end
end, value = "number;string", ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Modifiers.aggrocolor = {function(unit, value)
	local val = UnitReaction("player", unit) or 5

	local info = UnitReactionColor[val]
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	else
		return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	end
end, value = "number;string", ret = "string", events = "UNIT_FACTION" }

Modifiers.difficultycolor = {function(unit, value)
	if UnitCanAttack("player", unit) then
		local level = UnitLevel(unit)
		if level == 0 then
			level = 99
		end
		local info = GetDifficultyColor(level)
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
		else
			return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		end
	end
	return value
end, value = "number;string", ret = "string", events = "UNIT_LEVEL;PLAYER_LEVEL_UP" }

Modifiers.manacolor = {function(unit, value)
	local powerType = UnitPowerType(unit)
	local r,g,b
	if powerType == 0 then
		r,g,b = unpack(colors.mana)
	elseif powerType == 1 then
		r,g,b = unpack(colors.rage)
	elseif powerType == 2 then
		r,g,b = unpack(colors.focus)
	elseif powerType == 3 then
		r,g,b = unpack(colors.energy)
	else
		r,g,b = unpack(colors.unknown)
	end
	if value then
		return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	else
		return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end
end, value = "number;string", ret = "string", events = "UNIT_DISPLAYPOWER" }

Modifiers.combatcolor = {function(unit, value)
	if UnitAffectingCombat(unit) then
		local r,g,b = unpack(colors.inCombat)
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		else
			return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		end
	else
		if value then
			return ("|r%s"):format(value)
		else
			return "|r"
		end
	end
end, value = "number;string", ret = "string", events = "PLAYER_REGEN_ENABLED;PLAYER_REGEN_DISABLED;Update" }

Modifiers.color = {function(unit, arg, value)
	if value then
		return ("|cff%s%s|r"):format(arg, value)
	else
		return ("|cff%s"):format(arg)
	end
end, arg = "string", value = "number;string", ret = "string"}
-- colors
Modifiers.white = "color(ffffff)"
Modifiers.red = "color(ff0000)"
Modifiers.green = "color(00ff00)"
Modifiers.blue = "color(0000ff)"
Modifiers.cyan = "color(00ffff)"
Modifiers.fuchsia = "color(ff00ff)"
Modifiers.yellow = "color(ffff00)"
Modifiers.gray = "color(afafaf)"

local Tags = {}
-- When writing tags, `unit` is modified before the call, that contains the unit id.
-- Also, nil (which translates to empty string), a number, or a string can be returned.
-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- Note: can take an argument, either nil, a number, or a string.

Tags.isfriend = {function(unit)
	return UnitIsFriend("player", unit) and "true"
end, ret = "string;nil", events = "UNIT_FACTION" }

Tags.isenemy = "~isfriend"

Tags.text = {function(unit, arg)
	return arg
end, arg = "string", ret = "string"}

Tags.name = {UnitName, ret = "string", events = "UNIT_NAME_UPDATE"}

Tags.namehostile = "name:hostilecolor"

Tags.status = {function(unit)
	if not UnitIsConnected(unit) then
		return L["Offline"]
	elseif UnitIsDead(unit) then
		return L["Dead"]
	elseif UnitIsGhost(unit) then
		return L["Ghost"]
	elseif UnitAffectingCombat(unit) then
		return L["Combat"]
	elseif unit == "player" and IsResting() then
		return L["Resting"]
	end
end, ret = "string;nil", events = "Update" --[[due to the combat thing]]}

Tags.statusgone = {function(unit)
	if not UnitIsConnected(unit) then
		return L["Offline"]
	elseif UnitIsDead(unit) then
		return L["Dead"]
	elseif UnitIsGhost(unit) then
		return L["Ghost"]
	end
end, ret = "string;nil", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST" }

Tags.curhp = {function(unit)
	local currValue = UnitHealth(unit)
	local MobHealth3 = MobHealth3
	if MobHealth3 and not UnitIsFriend("player", unit) then
		local maxValue = UnitHealthMax(unit)
		local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
		if MHfound then
			return curr
		end
	end
	return currValue or 0
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.maxhp = {function(unit)
	local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
	if maxValue == 0 then
		return 0
	end
	local MobHealth3 = MobHealth3
	if MobHealth3 and not UnitIsFriend("player", unit) then
		local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue,maxValue)
		if MHfound then
			return max
		end
	end
	return maxValue
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.percenthp = {function(unit)
	local hpmax = UnitHealthMax(unit)
	if hpmax == 0 then
		return 0
	end
	return math.floor(UnitHealth(unit) / hpmax * 100 + 0.5)
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.missinghp = {function(unit)
	local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
	if maxValue == 0 then
		return 0
	end
	local MobHealth3 = MobHealth3
	if MobHealth3 and not UnitIsFriend("player", unit) then
		local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
		if MHfound then
			return max - curr
		end
	end
	return maxValue - currValue
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.fractionalhp = {function(unit)
	local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
	local MobHealth3 = MobHealth3
	if maxValue > 100 then
		return currValue.."/"..maxValue
	elseif MobHealth3 and not UnitIsFriend("player", unit) then
		local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
		if MHfound then
			return curr.."/"..max
		end
	end
	return currValue.."/"..maxValue
end, ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.smarthp = {function(unit)
	local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
	local MobHealth3 = MobHealth3
	if maxValue > 100 then
		return currValue.."/"..maxValue
	end
	if MobHealth3 and not UnitIsFriend("player", unit) then
		local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
		if MHfound then
			return curr.."/"..max
		end
	end
	return ("%.0f%%"):format(currValue / maxValue * 100)
end, ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH" }

Tags.curmana = {function(unit)
	return UnitMana(unit) or 0
end, ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Tags.maxmana = {function(unit)
	return UnitManaMax(unit)
end, ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Tags.percentmana = {function(unit)
	local mpmax = UnitManaMax(unit)
	if mpmax == 0 then
		return 0
	end
	return math.floor(UnitMana(unit) / mpmax * 100 + 0.5)
end, ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Tags.missingmana = {function(unit)
	return UnitManaMax(unit) - UnitMana(unit)
end, ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Tags.fractionalmana = {function(unit)
	return UnitMana(unit).."/"..UnitManaMax(unit)
end, ret = "string", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER" }

Tags.typemana = {function(unit)
	local p = UnitPowerType(unit)
	if p == 1 then
		return L["Rage"]
	elseif p == 2 then
		return L["Focus"]
	elseif p == 3 then
		return L["Energy"]
	else
		return L["Mana"]
	end
end, ret = "string", events = "UNIT_DISPLAYPOWER" }

Tags.level = {function(unit)
	local x = UnitLevel(unit)
	if x > 0 then
		return x
	else
		return "??"
	end
end, ret = "number;string", events = "UNIT_LEVEL" }

Tags.class = {function(unit)
	return UnitClass(unit) or UNKNOWN
end, ret = "string"}

Tags.smartclass = {function(unit)
	if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
		return UnitClass(unit)
	end
end, ret = "string;nil"}

Tags.playerclass = {function(unit)
	if UnitIsPlayer(unit) then
		return UnitClass(unit)
	end
end, ret = "string;nil"}

Tags.shortclass = {function(unit)
	local _,c = UnitClass(unit)
	if c then
		return shortclasses[c]
	end
end, ret = "string;nil"}

Tags.shortsmartclass = {function(unit)
	if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
		local _,c = UnitClass(unit)
		return shortclasses[c]
	end
end, ret = "string;nil"}

Tags.shortplayerclass = {function(unit)
	if UnitIsPlayer(unit) then
		local _,c = UnitClass(unit)
		return shortclasses[c]
	end
end, ret = "string;nil"}

Tags.creature = {function(unit)
	return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
end, ret = "string"}

Tags.combos = {function(unit)
	return GetComboPoints() or 0
end, ret = "number", events = "PLAYER_COMBO_POINTS" }

Tags.combosymbols = {function(unit)
	local num = GetComboPoints()
	if num >= 1 then
		return ("@"):rep(num)
	end
end, ret = "string;nil", events = "PLAYER_COMBO_POINTS" }

Tags.classification = {function(unit)
	if UnitClassification(unit) == "rare" then
		return L["Rare"]
	elseif UnitClassification(unit) == "rareelite" then
		return L["Rare-Elite"]
	elseif UnitClassification(unit) == "elite" then
		return L["Elite"]
	elseif UnitClassification(unit) == "worldboss" then
		return L["Boss"]
	end
end, ret = "string;nil"}

Tags.shortclassification = {function(unit)
	if UnitClassification(unit) == "rare" then
		return L["r"]
	elseif UnitClassification(unit) == "rareelite" then
		return L["r+"]
	elseif UnitClassification(unit) == "elite" then
		return L["+"]
	elseif UnitClassification(unit) == "worldboss" then
		return L["b"]
	end
end, ret = "string;nil"}

Tags.faction = {function(unit)
	local _, faction = UnitFactionGroup(unit)
	return faction
end, ret = "string"}

Tags.connect ={ function(unit)
	if not UnitIsConnected(unit) then
		return L["Offline"]
	end
end, ret = "string;nil"}

Tags.smartrace = {function(unit)
	if UnitIsPlayer(unit) then
		return UnitRace(unit) or UNKNOWN
	else
		return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
	end
end, ret = "string"}

Tags.race = {UnitRace, ret = "string"}

Tags.shortrace = {function(unit)
	local _,r = UnitRace(unit)
	return shortraces[r]
end, ret = "string;nil"}

Tags.pvp = {function(unit)
	return UnitIsPVP(unit) and L["PvP"]
end, ret = "string;nil"}

Tags.plus = {function(unit)
	return UnitIsPlusMob(unit) and "+"
end, ret = "string;nil"}

Tags.sex = {function(unit)
	local sex = UnitSex(unit)
	if sex == 0 then
		return L["Male"]
	elseif sex == 1 then
		return L["Female"]
	end
end, ret = "string;nil"}

Tags.shortsex = {function(unit)
	local sex = UnitSex(unit)
	if sex == 0 then
		return L["m"]
	elseif sex == 1 then
		return L["f"]
	end
end, ret = "string;nil"}

Tags.rested = {function(unit)
	if GetRestState() then
		return L["Rested"]
	end
end, ret = "string;nil", events = "PLAYER_XP_UPDATE" }

Tags.leader = {function(unit)
	if UnitIsPartyLeader(unit) then
		return L["L"]
	end
end, ret = "string;nil"}

Tags.leaderlong = {function(unit)
	if UnitIsPartyLeader(unit) then
		return L["Leader"]
	end
end, ret = "string;nil"}

Tags.happynum = {function(unit)
	return GetPetHappiness() or 0
end, ret = "number", events = "UNIT_HAPPINESS" }

Tags.happytext = {function(unit)
	return _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)]
end, ret = "string;nil", events = "UNIT_HAPPINESS" }

Tags.happyicon = {function(unit)
	local happy = GetPetHappiness()
	if happy == 3 then
		return ":D"
	elseif happy == 2 then
		return ":||"
	elseif happy == 1 then
		return "B("
	end
end, ret = "string;nil", events = "UNIT_HAPPINESS" }

Tags.curxp = {function(unit)
	if unit == "pet" then
		return (GetPetExperience())
	end
	return UnitXP(unit)
end, ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE" }

Tags.maxxp = {function(unit)
	if unit == "pet" then
		local _,xp = GetPetExperience()
		return xp
	end
	return UnitXPMax(unit)
end, ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE" }

Tags.percentxp = {function(unit)
	if unit == "pet" then
		local cur,max = GetPetExperience()
		if max > 0 then
			return math.floor(cur/xp * 100 + 0.5)
		else
			return 0
		end
	end
	local max = UnitXPMax(unit)
	if max > 0 then
		return math.floor(UnitXP(unit) / max * 100 + 0.5)
	end
	return 0
end, ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE" }

Tags.missingxp = {function(unit)
	if unit == "pet" then
		local cur, max = GetPetExperience()
		return max - cur
	end
	return UnitXPMax(unit) - UnitXP(unit)
end, ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE" }

Tags.restxp = {function(unit)
	return unit == "player" and GetXPExhaustion() or 0
end, ret = "number", "PLAYER_XP_UPDATE" }

Tags.percentrestxp = {function(unit)
	local rest = unit == "player" and GetXPExhaustion() or 0
	local max = UnitXPMax(unit)
	if max > 0 then
		return math.floor(rest / max * 100 + 0.5)
	end
	return 0
end, ret = "number", events = "PLAYER_XP_UPDATE" }

Tags.tappedbyme = {function(unit)
	if UnitIsTappedByPlayer(unit) then
		return "*"
	end
end, ret = "string;nil", events = "Update" }

Tags.istapped = {function(unit)
	if UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
		return L["Tapped"]
	end
end, ret = "string;nil", events = "Update" }

Tags.fkey = {function(unit)
	local fkey
	if unit == "player" then
		fkey = 0
	else
		fkey = unit:match("^party(%d)$")
	end
	if fkey then
		return "F" .. (fkey + 1)
	end
end, ret = "string;nil"}

Tags.raidgroup = {function(unit)
	for i = 1, GetNumRaidMembers() do
		local name, rank, subgroup = GetRaidRosterInfo(i)
		if name == UnitName(unit) then
			return subgroup
		end
	end
end, ret = "string"}

Tags.guild = {GetGuildInfo, ret = "string;nil"}

Tags.guildrank = {function(unit)
	local _, rank = GetGuildInfo(unit)
	return rank
end, ret = "string;nil"}

Tags.target = {function(unit)
	local u = unit .. "target"
	if UnitExists(u) then
		return UnitName(u)
	end
end, ret = "string;nil", events = "UNIT_TARGET" }

Tags.statuscolor = Modifiers.statuscolor
Tags.happycolor = Modifiers.happycolor
Tags.hostilecolor = Modifiers.hostilecolor
Tags.classcolor = Modifiers.classcolor
Tags.hpcolor = Modifiers.hpcolor
Tags.aggrocolor = Modifiers.aggrocolor
Tags.difficultycolor = Modifiers.difficultycolor
Tags.manacolor = Modifiers.manacolor
Tags.combatcolor = Modifiers.combatcolor
Tags.color = Modifiers.color
-- colors
Tags.white = Modifiers.white
Tags.red = Modifiers.red
Tags.green = Modifiers.green
Tags.blue = Modifiers.blue
Tags.cyan = Modifiers.cyan
Tags.fuchsia = Modifiers.fuchsia
Tags.yellow = Modifiers.yellow
Tags.gray = Modifiers.gray

Tags.smartmissinghp = "missinghp:hidezero:negate"

local function sortStringList(s)
	if not s then
		return nil
	end
	local list = new((";"):split(s))
	table.sort(list)
	local q = table.concat(list, ';')
	list = del(list)
	return q
end

local TagEvents = {}
local ModifierEvents = {}
local TagAliases = {}
local ModifierAliases = {}
local TagArguments = {}
local ModifierArguments = {}
local TagReturns = {}
local ModifierReturns = {}
local ModifierValues = {}
do
	for tag, data in pairs(Tags) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg then
				TagArguments[tag] = arg
				local a,b = (";"):split(arg)
				assert(a == "number" or a == "string")
				assert(not b or b == "number" or b == "string")
			end
			TagReturns[tag] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			TagEvents[tag] = sortStringList(data.events)
			local func = data[1]
			if Modifiers[tag] ~= Tags[tag] then
				del(Tags[tag])
			end
			Tags[tag] = func
		elseif type(data) == "string" then
			TagAliases[tag] = data
			Tags[tag] = nil
		end
	end
	
	for mod, data in pairs(Modifiers) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg then
				ModifierArguments[mod] = arg
				local a,b = (";"):split(arg)
				assert(a == "number" or a == "string")
				assert(not b or b == "number" or b == "string")
			end
			ModifierReturns[mod] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string" or a == "same")
				assert(not b or b == "nil" or b == "number" or b == "string" or b == "same")
				assert(not c or c == "nil" or c == "number" or c == "string" or c == "same")
			end
			local value = sortStringList(data.value)
			if value and value ~= "number;string" then
				ModifierValues[mod] = value
				local a,b = (";"):split(value)
				assert(a == "number" or a == "string")
				assert(not b or b == "number" or b == "string")
			end
			ModifierEvents[mod] = sortStringList(data.events)
			local func = data[1]
			del(Modifiers[mod])
			Modifiers[mod] = func
		elseif type(data) == "string" then
			ModifierAliases[mod] = data
			Modifiers[mod] = nil
		end
	end
end

local function toliteral(value)
	if type(value) == "number" then
		return ("%s"):format(value)
	else
		return ("%q"):format(value)
	end
end

local function splitCode(code)
	local alpha, bravo, charlie = code:match("^([^:!%?]*)([:!%?])(.*)$")
	if not alpha then
		return code
	end
	return alpha, bravo, splitCode(charlie)
end

local function splitParam(code)
	local negate, alpha, bravo = code:match("^(~*)([a-zA-Z]+)%(([^%)]+)%)$")
	if alpha then
		return alpha, bravo, ((negate:len() % 2) == 1 and true or false)
	else
		negate, alpha = code:match("^(~*)(.+)$")
		if alpha then
			return alpha, nil, ((negate:len() % 2) == 1 and true or false)
		else
			return code, nil, nil
		end
	end
end

local function _compileTree(code, funcs)
	if type(code) == "string" then
		return code, 0
	end
	local ends
	code[1], ends = _compileTree(code[1], funcs)
	if not code[1] then
		code = del(code)
		return nil, ends
	end
	local t = new()
	if not code[2] then
		local tag, param, negate = splitParam(code[1])
		
		local func = Tags[tag]
		if func then
			if param then
				if not TagArguments[tag] then
					t = del(t)
					code = del(code)
					return nil, ("Unexpected parameter, %q, to %q"):format(param, tag)
				end
				if TagArguments[tag]:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not TagArguments[tag]:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("Bad parameter, %q, to %q. Expected number."):format(param, tag)
					end
				end
			else
				if TagArguments[tag] then
					t = del(t)
					code = del(code)
					return nil, ("Expected parameter for %q"):format(tag)
				end
			end
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "not "
			end
			local tagID = "tag_" .. tag
			if not funcs[tagID] then
				funcs[tagID] = ("Tags[%q]"):format(tag)
			end
			t[#t+1] = tagID
			t[#t+1] = "(unit"
			if param then
				t[#t+1] = ", "
				t[#t+1] = toliteral(param)
			end
			t[#t+1] = ")"
			if negate then
				t[#t+1] = " and 'true'"
			end
			t[#t+1] = ";"
		elseif tag == "" then
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "'true'"
			else
				t[#t+1] = "nil"
			end
			t[#t+1] = ";"
		else
			t = del(t)
			code = del(code)
			return nil, ("Unknown tag %q"):format(tag)
		end
	elseif code[2] == "?" then
		t[#t+1] = code[1]
		t[#t+1] = ("if value then ")
		local q, e = _compileTree(code[3], funcs)
		if not q then
			t = del(t)
			code = del(code)
			return nil, e
		end
		t[#t+1] = q
		for i = 1, e do
			t[#t+1] = "end;"
		end
		if code[4] == "!" then
			t[#t+1] = ("else ")
			local q, e = _compileTree(code[5], funcs)
			if not q then
				t = del(t)
				code = del(code)
				return nil, e
			end
			t[#t+1] = q
			for i = 1, e do
				t[#t+1] = "end;"
			end
		end
		t[#t+1] = ("end;")
	elseif code[2] == ":" then
		t[#t+1] = code[1]
		local mod, param, negate = splitParam(code[3])
		local func = Modifiers[mod] or Tags[mod]
		if func then
			local isMod = func == Modifiers[mod]
			if param then
				if (isMod and not ModifierArguments[mod]) or (not isMod and not TagArguments[mod]) then
					t = del(t)
					code = del(code)
					return nil, ("Unexpected parameter, %q, to %q"):format(param, mod)
				end
				
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if arg:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not arg:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("Bad parameter, %q, to %q. Expected number."):format(param, mod)
					end
				end
			else
				if (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod]) then
					t = del(t)
					code = del(code)
					return nil, ("Expected parameter for %q"):format(mod)
				end
			end
			if isMod then
				local val = ModifierValues[mod]
				if not val then
					t[#t+1] = "if value then "
				else
					t[#t+1] = [[if type(value) == "]]
					t[#t+1] = val
					t[#t+1] = [[" then ]]
				end
			else
				t[#t+1] = "if not value then "
			end
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "not "
			end
			local f
			if isMod then
				local modID = "mod_" .. mod
				if not funcs[modID] then
					funcs[modID] = ("Modifiers[%q]"):format(mod)
				end
				f = modID
			else
				local tagID = "tag_" .. mod
				if not funcs[tagID] then
					funcs[tagID] = ("Tags[%q]"):format(mod)
				end
				f = tagID
			end
			t[#t+1] = f
			t[#t+1] = "(unit"
			if param then
				t[#t+1] = ", "
				t[#t+1] = toliteral(param)
			end
			if isMod then
				t[#t+1] = ", "
				t[#t+1] = "value"
			end
			t[#t+1] = ")"
			if isMod then
				if negate then
					t[#t+1] = " and value"
				end
				t[#t+1] = " end;"
			else
				if negate then
					t[#t+1] = " and 'true'"
				end
				t[#t+1] = ";"
				ends = ends + 1
			end
		else
			t = del(t)
			code = del(code)
			return nil, ("Unknown modifier or tag %q"):format(mod)
		end	
	else
		t = del(t)
		code = del(code)
		return nil, ("Unknown separator %q"):format(code[2])
	end
	code = del(code)
	local s = table.concat(t)
	t = del(t)
	return s, ends
end

local function compileTree(code, funcs)
	local q, e = _compileTree(code, funcs)
	if not q then
		return "value = [===[[" .. e .. "]]===];"
	end
	for i = 1, e do
		q = q .. "end;"
	end
	return q
end

local function select2(min, max, ...)
	if min <= max then
		return select(min, ...), select2(min+1, max, ...)
	end
end

local function parse(...)
	local n = select('#', ...)
	if n == 1 then
		return new((...))
	end
	local q_num = 0
	local q_pos
	local x_pos
	for i = 1, n do
		if select(i, ...) == "?" then
			q_num = q_num + 1
			if not q_pos then
				q_pos = i
			end
		elseif select(i, ...) == "!" then
			q_num = q_num - 1
			if q_num == 0 then
				x_pos = i
				break
			end
		end
	end
	if q_pos then
		if not x_pos then
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select(q_pos + 1, ...)))
		else
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select2(q_pos + 1, x_pos - 1, ...)), "!", parse(select(x_pos + 1, ...)))
		end
	end
	return new(parse(select2(1, n - 2, ...)), ":", select(n, ...))
end

local tmp = {}
local codeToFunction = setmetatable({[""] = function() return "" end}, {__index=function(self, code)
	local t = new()
	t[#t+1] = ([[local DogTag = AceLibrary(%q);]]):format(MAJOR_VERSION)
	t[#t+1] = [[local Tags = DogTag.__Tags; local Modifiers = DogTag.__Modifiers;]]
	t[#t+1] = "" -- placeholder
	t[#t+1] = [[local table_concat = table.concat;]]
	t[#t+1] = [[local string_gsub = string.gsub;]]
	t[#t+1] = [[local type = type;]]
	t[#t+1] = [[local tonumber = tonumber;]]
	t[#t+1] = [[local tmp = DogTag.__tmp;]]
	t[#t+1] = [[return function(unit) ]]
	t[#t+1] = [[if not UnitExists(unit) then return ]]
	if code:find("%[name[%]:%?!]") then
		t[#t+1] = [[unit]]
	else
		t[#t+1] = [[""]]
	end
	t[#t+1] = [[ end;]]
	local funcs = new()
	local len = code:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = code:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			local rest = code:sub(st)
			if rest:len() > 0 then
				t[#t+1] = ("tmp[#tmp+1] = %q;"):format(rest)
			end
			break
		end
		st = fin+1
		if left:len() > 0 then
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(left)
		end
		
		local u = new()
		u[#u+1] = "do local value;"
		
		local result = compileTree(parse(splitCode(inner)), funcs)
		if not result then
			u = del(u)
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(literal)
		else
			u[#u+1] = result
			u[#u+1] = [[if value then tmp[#tmp+1] = value end;]]
			u[#u+1] = [[end;]]
			local s = table.concat(u)
			u = del(u)
			t[#t+1] = s
		end
	end
	local u = new()
	for k, v in pairs(funcs) do
		u[#u+1] = "local "
		u[#u+1] = k
		u[#u+1] = " = "
		u[#u+1] = v
		u[#u+1] = ";"
	end
	t[3] = table.concat(u)
	u = del(u)
	t[#t+1] = [[local s = table_concat(tmp);]]
	t[#t+1] = [[for i = 1, #tmp do tmp[i] = nil end;]]
	t[#t+1] = [[return (string_gsub(string_gsub(string_gsub(s, "^ +", ""), " +$", ""), "  +", " "));]]
	t[#t+1] = [[end;]]
	local s = table.concat(t)
	t = del(t)
	DogTag.__Tags = Tags
	DogTag.__Modifiers = Modifiers
	DogTag.__tmp = tmp
	local val = assert(loadstring(s))()
	DogTag.__Tags = nil
	DogTag.__Modifiers = nil
	DogTag.__tmp = nil
	self[code] = val
	return val
end, __mode='k'})

local codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	local len = key:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = key:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			break
		end
		st = fin+1
		local code = new(splitCode(inner))
		for i = 1, #code, 2 do
			local bit, param = splitParam(code[i])
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			local isMod = func == Modifiers[bit]
			local eventList = (isMod and ModifierEvents[bit]) or (not isMod and TagEvents[bit])
			if eventList then
				local events = new((";"):split(eventList))
				for _,event in ipairs(events) do
					t[event] = true
				end
				events = del(events)
			end
		end
		code = del(code)
	end
	self[key] = t
	return t
end, __mode = 'k'})

local antiAlias
do
	local function handler(literal, inner)
		local code = new(splitCode(inner))
		local real = new()
		for i = 1, #code, 2 do
			local bit, param, negate = splitParam(code[i])
			bit = bit:lower()
			local func
			if code[i-1] ~= ":" then
				func = TagAliases[bit]
			else
				func = ModifierAliases[bit]
				if not func and not Modifiers[bit] then
					func = TagAliases[bit]
				end
			end
			if i == 1 then
				real[1] = "["
			else
				real[#real+1] = code[i-1]
			end
			if negate then
				real[#real+1] = "~"
			end
			if func then
				if param and func:find("[:%(%?!]") then
					if func:find("%(arg%)") then
						real[#real+1] = func:gsub("%(arg%)", "(" .. param .. ")")
					else
						real[#real+1] = bit
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				else
					real[#real+1] = func
					if param then
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				end
			else
				real[#real+1] = bit
				if param then
					real[#real+1] = "("
					real[#real+1] = param
					real[#real+1] = ")"
				end
			end
		end
		code = del(code)
		
		real[#real+1] = ']'
		local str = table.concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		local value = key
		while true do
			local lastValue = value
			value = lastValue:gsub("(%[([^%]]+)%])", handler)
			if value == lastValue then
				break
			end
		end
		self[key] = value
		return value
	end, __mode= 'k'})
end

local eventData = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	if event ~= "Update" then
		DogTag.frame:RegisterEvent(event)
	end
	return t
end})
local buckets = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	return t
end})

local registry
local codeHash

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})

local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end
function DogTag:RegisterFontString(text, unit, code)
	self:argCheck(text, 2, "table")
	self:argCheck(unit, 3, "string")
	self:argCheck(code, 4, "string")
	if not IsLegitimateUnit[unit] then
		code = ""
	end
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == unit then
			if text:IsShown() then
				text:SetText(codeToFunction[code](unit))
			end
			return
		end
		self:UnregisterFontString(text)
	end
	codeHash[text] = code
	registry[text] = unit
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			eventData[event][text] = true
		end
	end
	registry[text] = unit
	
	if text:IsShown() then
		text:SetText(codeToFunction[code](unit))
	end
end

function DogTag:UnregisterFontString(text)
	self:argCheck(text, 2, "table")
	local code = codeHash[text]
	if not code then
		return
	end
	
	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event ~= "Update" then
					self.frame:UnregisterEvent(event)
					buckets[event] = del(buckets[event])
				end
			end
		end
	end
	
	text:SetText("")
end

function DogTag:UpdateFontString(text)
	self:argCheck(text, 2, "table")
	local unit = registry[text]
	if unit and text:IsShown() then
		text:SetText(codeToFunction[codeHash[text]](unit))
	end
end

function DogTag:UpdateAllForUnit(unit)
	self:argCheck(unit, 2, "string")
	if not IsLegitimateUnit[unit] then
		return
	end
	
	for text, u in pairs(registry) do
		if u == unit and text:IsShown() then
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
end

function DogTag:Evaluate(unit, code)
	if not IsLegitimateUnit[unit] then
		return ""
	end
	return codeToFunction[antiAlias[code]](unit)
end

function DogTag:PARTY_MEMBERS_CHANGED()
	for text, unit in pairs(registry) do
		if text:IsShown() then
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
end
DogTag.PARTY_LEADER_CHANGED = DogTag.PARTY_MEMBERS_CHANGED
DogTag.RAID_ROSTER_UPDATE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_ENABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_DISABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_ENTERING_WORLD = DogTag.PARTY_MEMBERS_CHANGED
DogTag.UNIT_PET = DogTag.PARTY_MEMBERS_CHANGED

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:UPDATE_MOUSEOVER_UNIT()
	self:UpdateAllForUnit('mouseover')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

function DogTag:SetColorConstantTable(t)
	colors = t
	self:PARTY_MEMBERS_CHANGED()
end

local function activate(self, oldLib, oldDeactivate)
	DogTag = self
	
	self.frame = oldLib and oldLib.frame or CreateFrame("Frame", "DogTag10Frame")
	
	self.frame:UnregisterAllEvents()
	self.frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self.frame:RegisterEvent("RAID_ROSTER_UPDATE")
	self.frame:RegisterEvent("PARTY_MEMBER_ENABLE")
	self.frame:RegisterEvent("PARTY_MEMBER_DISABLE")
	self.frame:RegisterEvent("PARTY_ENTERING_WORLD")
	self.frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	self.frame:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self.frame:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self.frame:RegisterEvent("PLAYER_PET_CHANGED")
	self.frame:RegisterEvent("UNIT_PET")
	self.frame:SetScript("OnEvent", function(this, event, arg1)
		local func = self[event]
		if func then
			func(self)
			return
		end
		buckets[event][arg1 or false] = true
	end)
	local timePassed = 0
	local num = 0
	local toUpdate = {}
	self.frame:SetScript("OnUpdate", function(this, elapsed)
		timePassed = timePassed + elapsed
		if timePassed > 0.05 then
			timePassed = 0
			num = (num + 1) % 3
			if num == 0 then
				if eventData.Update then
					for text in pairs(eventData.Update) do
						toUpdate[text] = true
					end
				end
			end
			for event, units in pairs(buckets) do
				if units[false] then
					for text in pairs(eventData[event]) do
						toUpdate[text] = true
					end
				else
					for text in pairs(eventData[event]) do
						if units[registry[text]] then
							toUpdate[text] = true
						end
					end
				end
			end
			for text in pairs(toUpdate) do
				toUpdate[text] = nil
				local unit = registry[text]
				if unit and text:IsShown() then
					text:SetText(codeToFunction[codeHash[text]](unit))
				end
			end
		end
	end)
	
	registry = {}
	self.registry = registry
	codeHash = {}
	self.codeHash = codeHash
	if oldLib then
		for text, unit in pairs(oldLib.registry) do
			self:RegisterFontString(text, unit, oldLib.codeHash[text])
		end
	end
end

AceLibrary:Register(DogTag, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
